<?php

declare(strict_types=1);

namespace app\models;

use yii;
use yii\web\User as Account;

use app\models\traits\SearchByEdge;

use Exception;

/**
 * Заказ
 *
 * @see Account Заказчик
 * @see Supply Поставки для заказа
 */
class Order extends Document
{
    use SearchByEdge;

    /**
     * Поставки для записи
     */
    public array $supplies;

    /**
     * Имя коллекции
     */
    public static function collectionName(): string
    {
        return 'order';
    }

    /**
     * Свойства
     */
    public function attributes(): array
    {
        return array_merge(
            parent::attributes(),
            []
        );
    }

    /**
     * Метки свойств
     */
    public function attributeLabels(): array
    {
        return array_merge(
            parent::attributeLabels(),
            []
        );
    }

    /**
     * Правила
     */
    public function rules(): array
    {
        return array_merge(
            parent::rules(),
            []
        );
    }

    /**
     * Запись
     *
     * $supply = [ Supply $supply, int $amount = 1 ]
     *
     * @param Supply|array $supply Поставка
     * @param Account $trgt Заказчик
     *
     * @todo Создать параметр разделителя для администрации
     */
    public function write(Supply|array $supply, Account $trgt = null): self|null
    {
        // Инициализация
        $trgt ?? $trgt = yii::$app->user ?? throw new Exception('Не удалось инициализировать заказчика');

        if ($supply instanceof Supply) {
            // Передана инстанция класса поставки или второй элемент массива не является числом

            // Унификация входных данных
            $supply = [$supply, 1];
        }

        if (is_null($this->_key)) {
            // Корзина не инициализирована

            // Инициализация
            if (!$this->save()) {
                // Инициализация заказа не удалась

                throw new Exception('Ошибка при записи заказа в базу данных');
            }

            // Инициализация ребра: АККАУНТ -> ЗАКАЗ
            if (!AccountEdgeOrder::write($trgt->readId(), $this->readId(), 'create')) {
                // Инициализация не удалась

                throw new Exception('Ошибка при записи ребра от аккаунта до заказа в базу данных');
            }
        }

        foreach (is_array($supply[0]) ? $supply : [$supply] as $supply_raw) {
            // Перебор поставок

            for ($i = 0; $i < $supply_raw[1]; $i++) {
                // Создание рёбер соразмерно запросу (добавление нескольких продуктов в корзину)

                // Запись ребра: ЗАКАЗ -> ПОСТАВКА
                if (!$supply = Supply::searchByCatn($supply_raw[0]) or !OrderEdgeSupply::write($this->readId(), $supply->readId(), 'write')) {
                    // Поставка не найдена или запись ребра не удалась

                    continue;
                }
            }
        }

        // Отправка на сервер базы данных
        return $this->save() ? $this : null;
    }

    /**
     * Подключение к аккаунту
     */
    public function connect(Account $account): ?AccountEdgeOrder {
        // Запись ребра: АККАУНТ -> ЗАКАЗ
        return AccountEdgeOrder::write($account->id, $this->readId(), 'current') ?? throw new Exception('Не удалось инициализировать ребро: АККАУНТ -> ЗАКАЗ');
    }

    /**
     * Удаление
     */
    public function deleteEdge(string|array $catn): ?OrderEdgeSupply
    {
        // Запись ребра: ЗАКАЗ -> ПОСТАВКА
        return OrderEdgeSupply::write($this->readId(), Supply::searchByCatn($catn)->readId(), 'delete');
    }

    /**
     * Поиск заказа
     */
    public static function search(Account $account = null, string $type = 'current', int $limit = 1, int $page = 1, string $select = null): self|array|null
    {
        // Инициализация
        $account or $account = yii::$app->user ?? throw new Exception('Не удалось инициализировать пользователя');

        // Генерация сдвига по запрашиваемым данным (пагинация)
        $offset = $limit * ($page - 1);

        if (strcasecmp($type, 'all') !== 0) {
            // Если не указан параметр поиска всех заказов

            $where_type = [
                'account_edge_order.type' => $type
            ];
        } else {
            $where_type = [];
        }

        $return = self::searchByEdge(
            from: 'account',
            to: 'order',
            subquery_where: [
                [
                    'account._id' => $account->id
                ],
                $where_type
            ],
            foreach: ['edge' => 'account_edge_order'],
            where: 'edge._to == order._id',
            limit: $limit,
            offset: $offset,
            sort: ['DESC'],
            select: $select,
            direction: 'INBOUND'
        );

        return $limit === 1 ? $return[0] ?? null : $return;
    }

    /**
     * Поиск содержимого заказа
     *
     * @todo В будущем возможно заказ не только поставок реализовать
     */
    public function content(int $limit = 1, int $page = 1): Supply|array|null
    {
        // Генерация сдвига по запрашиваемым данным (пагинация)
        $offset = $limit * ($page - 1);


        // Подзапрос для проверки статуса уведомления относительно пользователя
        // Поиск рёбер: ПОЛЬЗОВАТЕЛЬ -> УВЕДОМЛЕНИЕ
        $let = Supply::find()
            ->for(['order', 'order_edge_supply_deleted'])
            ->traversal('supply', 'INBOUND')
            ->in('order_edge_supply')
            ->where([
                [
                    'order._id' => $this->readId()
                ],
                [
                    'order_edge_supply_deleted.type' => 'delete'
                ]
            ])
            ->select('order_edge_supply_deleted');

        // Генерация
        $let = $let->createCommand();

        // Поиск рёбер: ЗАКАЗ -> ПОСТАВКА
        $supplies = Supply::searchByEdge(
            from: 'order',
            to: 'supply',
            edge: 'order_edge_supply',
            subquery_where: [
                [
                    'order._id' => $this->readId()
                ],
                [
                    'order_edge_supply.type' => 'write'
                ]
            ],
            params: $let->getBindVars(),
            let: [
                'order_edge_supply_deleted',
                '(' . (string) $let . ')'
            ],
            foreach: ['edge' => 'order_edge_supply'],
            where: 'edge._to == supply._id && edge._to != order_edge_supply_deleted[0]._to',
            limit: $limit,
            offset: $offset,
            direction: 'INBOUND'
        );

        // Инициализация реестра дубликатов
        $registry = [];

        // Подсчёт и перестройка массива для очистки от дубликатов
        foreach ($supplies as $key => &$supply) {
            // Перебор поставок

            if (in_array($supply->catn, $registry)) {
                // Если данная поставка найдена в реестре

                // Удаление
                unset($supplies[$key]);

                // Пропуск итерации
                continue;
            }

            $amount = 0;

            // Повторный перебор для поиска дубликатов
            foreach ($supplies as &$supply4check) {
                if ($supply == $supply4check) {
                    // Найден дубликат

                    // Постинкрементация счётчика
                    $amount++;

                    // Запись в реестр
                    $registry[] = $supply4check->catn;
                }
            }

            // Запись количества для заказа
            $supply->amnt = $amount;
        }

        // Поиск стоимости для каждой поставки
        foreach ($supplies as $key => &$supply) {
            // Перебор поставок

            // Чтение стоимости
            $cost = $supply->readCost();

            if ($cost < 1) {
                // Если стоимость равна нулю (явная ошибка)

                $this->deleteEdge($supply->readId());

                // Удаление
                unset($supplies[$key]);

                // Пропуск итерации
                continue;
            }

            // Запись цены
            $supply->cost = $cost['ЦенаЗаЕдиницу'] . ' ' . $cost['Валюта'];
        }

        return $supplies;
    }
}
